from ..common import Queue, StandardComponent
from .driver import Driver
from .othread import Othread


class Backend(StandardComponent):
    def process(self, data):
        raise NotImplementedError


class BackendDriver(Driver):

    def setup(self):
        if hasattr(self._runtime_context, 'transport_thread'):
            self._queue = self._runtime_context.backend_thread_queue
        else:
            self._queue = self._runtime_context.frontend_thread_queue
        super(BackendDriver, self).setup()

    @property
    def queue(self):
        return self._queue

    def run(self):
        while True:
            if self._current_thread.stopping():
                if self.queue.qsize() <= 0:
                    if not hasattr(self._runtime_context, 'transport_thread'):
                        break
                    elif self._runtime_context.transport_thread.stopped():
                        break

            try:
                data = self.queue.get(block=True, timeout=0.1)
                self._component.process(data)
            except Queue.Empty:
                pass


class BackendThread(Othread):

    def run(self):
        super(BackendThread, self).run()
        self._runtime_context.frontend_thread.stop()
        if hasattr(self._runtime_context, 'transport_thread'):
            self._runtime_context.transport_thread.stop()



import logging

from .common import binary_stdin
from .core.backend import Backend
from .core.frontend import Frontend
from .core.transport import Transport


class StdinFrontend(Frontend):
    def produce(self):
        for line in binary_stdin:
            yield line


class LogTransport(Transport):
    def setup(self):
        self._logger = logging.getLogger(self.__class__.__name__)

    def transport(self, data):
        self._logger.info(data)
        return data


class LogBackend(Backend):
    def setup(self):
        self._logger = logging.getLogger(self.__class__.__name__)

    def process(self, data):
        self._logger.info(data)



from ..common import Queue, StandardComponent
from .driver import Driver
from .othread import Othread


class Frontend(StandardComponent):

    def produce(self):
        raise NotImplementedError


class FrontendDriver(Driver):

    @property
    def queue(self):
        return self._runtime_context.frontend_thread_queue

    def run(self):
        for data in self._component.produce():
            while True:
                try:
                    self.queue.put(data, block=False, timeout=1)
                    break
                except Queue.Full:
                    continue
            if self._current_thread.stopping():
                break


class FrontendThread(Othread):

    def run(self):
        super(FrontendThread, self).run()
        if hasattr(self._runtime_context, 'transport_thread'):
            self._runtime_context.transport_thread.stop()
        elif hasattr(self._runtime_context, 'backend_thread'):
            self._runtime_context.backend_thread.stop()



import inspect
from importlib import import_module

import six


def getatters(obj, attrs):
    return [getattr(obj, attr) for attr in attrs if hasattr(obj, attr)]


def load_class(class_path, base_class=None, include_base_class=False):
    if isinstance(class_path,  six.string_types):
        module_path, class_name = class_path.rsplit('.', 1)
        _mod = import_module(module_path)
        if not hasattr(_mod, class_name):
            return None
        _cls = getattr(_mod, class_name)
        if not inspect.isclass(_cls):
            return None
    elif inspect.isclass(class_path):
        _cls = class_path

    if base_class is None:
        return _cls
    elif issubclass(_cls, base_class) and (include_base_class or _cls != base_class):
        return _cls
    else:
        return None



import pytest


def test_getatters():
    from os_m3_engine.utils import getatters

    class T(object):
        pass
    t = T()
    t.a = 1
    assert [1] == getatters(t, ['a'])


def test_load_class():
    from os_m3_engine.utils import load_class
    from os_m3_engine.core.engine import Engine
    assert Engine == load_class(
        'os_m3_engine.core.engine.Engine', Engine, True)

    from os_m3_engine.common import Configurable
    from os_m3_engine.common import Component
    assert Component == load_class('os_m3_engine.common.Component')

    assert None == load_class('os_m3_engine.NotExist')

    assert None == load_class('os_m3_engine.common.Component', Engine)

    with pytest.raises(ImportError):
        load_class('not_exist.What')



import sys

from .utils import load_class

_PY3 = sys.version_info[0] == 3

if _PY3:
    import queue as Queue
    binary_stdin = sys.stdin.buffer
else:
    import Queue
    if sys.platform == "win32":
        # set sys.stdin to binary mode
        import os
        import msvcrt
        msvcrt.setmode(sys.stdin.fileno(), os.O_BINARY)
    binary_stdin = sys.stdin


class Configurable(object):
    def __init__(self, config):
        self._config = config

    @property
    def config(self):
        return self._config


class Workflowable(object):
    def setup(self):
        pass

    def cleanup(self):
        pass


class Component(Configurable, Workflowable):
    pass


class StandardComponent(Component):
    def __init__(self, component_config, engine_config, runtime_context, current_thread):
        super(StandardComponent, self).__init__(component_config)
        self._engine_config = engine_config
        self._runtime_context = runtime_context
        self._current_thread = current_thread


class FactoryInterface(object):
    def create(self, *args, **kwargs):
        raise NotImplementedError


class ConfigurableFactory(Configurable):
    def create(self, *args, **kwargs):
        raise NotImplementedError


class ComponentFactory(ConfigurableFactory):

    def __init__(self, config, component_cls_string):
        super(ComponentFactory, self).__init__(config)
        self._component_cls = load_class(
            component_cls_string, StandardComponent)

    def create(self, engine_config, runtime_context, current_thread):
        return self._component_cls(self.config, engine_config,
                                   runtime_context, current_thread)



import threading
import time

from ..common import Configurable
from ..utils import getatters


class RuntimeContext(object):
    pass


class Engine(Configurable):

    def __init__(self, config, runtime_context):
        super(Engine, self).__init__(config)
        self.__start_lock = threading.Lock()
        self.__started = False
        self.__stopped = False
        self._runtime_context = runtime_context

    def started(self):
        return self.__started

    def stopped(self):
        return self.__stopped

    def start(self):

        self.__acquire_start_lock(False, 'Can not start twice')
        runtime_context = self._runtime_context
        m = getatters(runtime_context, [
                      'backend_thread',
                      'transport_thread',
                      'frontend_thread'
                      ])
        list(map(lambda x: x.start(), m))

        self.__started = True
        self.__start_lock.release()

        while not any([x.stopped() for x in m]):
            time.sleep(self.config.main_loop_wait)

        m = getatters(runtime_context, [
            'transport_thread',
            'backend_thread'
        ])
        list(map(lambda x: x.join(), m))
        self.__stopped = True

    def __acquire_start_lock(self, started, err_msg):
        self.__start_lock.acquire()
        if self.__started != started:
            try:
                raise RuntimeError(err_msg)
            finally:
                self.__start_lock.release()

    def __stop(self):
        runtime_context = self._runtime_context
        runtime_context.frontend_thread.stop()
        if hasattr(runtime_context, 'transport_thread'):
            runtime_context.transport_thread.stop()
        elif hasattr(runtime_context, 'backend_thread'):
            runtime_context.backend_thread.stop()
        self.__stopped = True
        self.__started = False

    def stop(self):
        self.__acquire_start_lock(True, 'Can stop before start')
        try:
            if not self.__stopped:
                self.__stop()
        finally:
            self.__start_lock.release()



import logging
import threading

from ..common import Configurable
from ..utils import load_class
from .driver import Driver


class Othread(Configurable, threading.Thread):

    def __init__(self, engine_config, runtime_context,
                 component_factory, start_trigger, stop_trigger, **kwargs):
        Configurable.__init__(self, engine_config)
        threading.Thread.__init__(self, **kwargs)
        self._driver_cls = load_class(engine_config.driver_cls, Driver)
        self._component_factory = component_factory
        self._runtime_context = runtime_context
        self._stop_trigger = stop_trigger
        self._start_trigger = start_trigger
        self._logger = logging.getLogger(self.name)

    def run(self):
        driver = None
        try:
            self._start_trigger.set()
            self._logger.debug('Start')
            driver = self._driver_cls(
                self.config,
                self._runtime_context,
                self,
                self._component_factory)
            driver.setup()
            driver.run()
        except Exception as e:
            self._logger.error('Unexpected exception, %s' % str(e))
        finally:
            try:
                if driver and hasattr(driver, 'cleanup'):
                    driver.cleanup()
            except Exception as e:
                self._logger.error('Unexpected exception, %s' % str(e))
        self._stop_trigger.set()
        self._logger.debug('Stop')

    def stopping(self):
        return self._stop_trigger.is_set()


class OthreadManager(Configurable):
    def __init__(self, engine_config, runtime_context, component_factory):
        super(OthreadManager, self).__init__(engine_config)
        start_trigger = threading.Event()
        stop_trigger = threading.Event()
        self._start_trigger = start_trigger
        self._stop_trigger = stop_trigger
        thread_cls = load_class(engine_config.thread_cls, Othread)
        self._threads = [thread_cls(
            engine_config, runtime_context, component_factory,
            start_trigger, stop_trigger,
            name='%s-%d' % (thread_cls.__name__, i))
            for i in range(0, engine_config.thread_num)]

    @property
    def thread_num(self):
        return len(self._threads)

    def start(self):
        list(map(lambda t: t.start(), self._threads))
        while not self._start_trigger.wait(0.1):
            if self._start_trigger.is_set():
                break
            

    def setDaemon(self, daemonic):
        list(map(lambda t: t.setDaemon(daemonic), self._threads))

    def join(self):
        list(map(lambda t: t.join(), self._threads))

    def stop(self):
        self._stop_trigger.set()

    def stopped(self):
        return not any([t.isAlive() for t in self._threads])



from ..common import Queue, StandardComponent
from .driver import Driver
from .othread import Othread


class Transport(StandardComponent):
    def transport(self, data):
        raise NotImplementedError


class TransportDriver(Driver):

    @property
    def input_queue(self):
        return self._runtime_context.frontend_thread_queue

    def _process(self, data):
        self._component.transport(data)

    def run(self):
        while True:
            if self._current_thread.stopping():
                if self.input_queue.qsize() <= 0:
                    break
                elif hasattr(self._runtime_context, 'backend_thread'):
                    if self._runtime_context.backend_thread.stopped():
                        break

            try:
                data = self.input_queue.get(block=True, timeout=0.1)
                self._process(data)
            except Queue.Empty:
                pass


class BridgeDriver(TransportDriver):

    @property
    def output_queue(self):
        return self._runtime_context.backend_thread_queue

    def _process(self, data):
        p_data = self._component.transport(data)
        while True:
            try:
                self.output_queue.put(p_data, block=True, timeout=0.1)
                break
            except Queue.Full:
                pass


class TransportThread(Othread):

    def run(self):
        super(TransportThread, self).run()
        self._runtime_context.frontend_thread.stop()
        if hasattr(self._runtime_context, 'backend_thread'):
            self._runtime_context.backend_thread.stop()



from ..common import Configurable, Workflowable


class Driver(Configurable, Workflowable):

    def __init__(self, engine_config, runtime_context, current_thread, component_factory):
        super(Driver, self).__init__(engine_config)
        self._runtime_context = runtime_context
        self._current_thread = current_thread
        self._component = component_factory.create(
            engine_config, runtime_context, current_thread)

    def setup(self):
        self._component.setup()

    def cleanup(self):
        self._component.cleanup()

    def run(self):
        raise NotImplementedError



import threading
import time

import pytest

from os_m3_engine.launcher import create


@pytest.fixture(scope='function')
def config():
    from os_config import Config
    return Config.create()


def test_engine_001(config):

    engine = create(
        frontend_cls='misc.OneRecordFrontend',
        transport_cls='misc.ConfirmTransport',
        backend_cls='misc.ConfirmBackend',
    )
    engine.start()


def test_engine_002(config):
    engine = create(
        frontend_cls='misc.SleepFrontend',
        transport_cls='misc.ConfirmTransport',
        backend_cls='misc.ConfirmBackend',
    )
    t1 = threading.Thread(target=engine.start)
    t1.start()

    while not engine.started():
        time.sleep(0.1)

    engine.stop()
    t1.join()
    assert engine.stopped()


def test_engine_003(config):
    from misc import OneRecordFrontend

    engine = create(
        frontend_cls=OneRecordFrontend,
        transport_cls='misc.ConfirmTransport',
        backend_cls='misc.ConfirmBackend',
    )
    engine.start()



import time
from os_m3_engine.core.backend import Backend
from os_m3_engine.core.frontend import Frontend
from os_m3_engine.core.transport import Transport

DATA = 'Hello World!'


class SleepFrontend(Frontend):
    def produce(self):
        time.sleep(100)
        yield DATA


class OneRecordFrontend(Frontend):
    def produce(self):
        assert 1 == 1
        yield DATA


class ConfirmTransport(Transport):
    def transport(self, data):
        assert data == DATA
        return data


class ConfirmBackend(Backend):
    def process(self, data):
        assert data == DATA



from os_config import Config

from .common import ConfigurableFactory, Queue
from .core.engine import Engine, RuntimeContext
from .core.othread import OthreadManager
from .utils import load_class

ENGINE_FRONTEND_CONFIG = Config.create(
    thread_cls='os_m3_engine.core.frontend.FrontendThread',
    thread_num=1,
    driver_cls='os_m3_engine.core.frontend.FrontendDriver',
    component_factory_cls='os_m3_engine.common.ComponentFactory',
    queue_size=100,
)

ENGINE_TRANSPORT_CONFIG = Config.create(
    thread_cls='os_m3_engine.core.transport.TransportThread',
    thread_num=3,
    driver_cls='os_m3_engine.core.transport.TransportDriver',
    component_factory_cls='os_m3_engine.common.ComponentFactory',
)

ENGINE_TRANSPORT_BRIDGE_CONFIG = Config.create()
ENGINE_TRANSPORT_BRIDGE_CONFIG.update(ENGINE_TRANSPORT_CONFIG)
ENGINE_TRANSPORT_BRIDGE_CONFIG.driver_cls = 'os_m3_engine.core.transport.BridgeDriver'

ENGINE_BACKEND_CONFIG = Config.create(
    thread_cls='os_m3_engine.core.backend.BackendThread',
    thread_num=1,
    driver_cls='os_m3_engine.core.backend.BackendDriver',
    component_factory_cls='os_m3_engine.common.ComponentFactory',
    queue_size=100,
)

ENGINE_CONFIG = Config.create(
    main_loop_wait=0.1
)


def _queue_size(custorm_config, default_config, thread_num):
    queue_size = custorm_config.queue_size if hasattr(
        custorm_config, 'queue_size') else default_config.queue_size
    if custorm_config == default_config \
            or not hasattr(custorm_config, 'queue_size'):
        queue_size = max(queue_size, thread_num*2)
    return queue_size


def combine_with_default_config(default_config, custom_config):
    c = Config.create()
    c.update(default_config)
    if custom_config is not None:
        for k, _ in default_config:
            if hasattr(custom_config, k):
                setattr(c, k, getattr(custom_config, k))
    return c


def create(frontend_cls='os_m3_engine.ootb.StdinFrontend',
           transport_cls='os_m3_engine.ootb.LogTransport',
           backend_cls='os_m3_engine.ootb.LogBackend',
           app_config=None,
           engine_config=ENGINE_CONFIG,
           engine_frontend_config=ENGINE_FRONTEND_CONFIG,
           engine_transport_config=ENGINE_TRANSPORT_BRIDGE_CONFIG,
           engine_backend_config=ENGINE_BACKEND_CONFIG,
           runtime_context=None):

    if frontend_cls is None:
        raise ValueError('Not specifiy frontend_cls')

    if transport_cls is None and backend_cls is None:
        raise ValueError('Spiecify at least one of transport_cls/backend_cls')

    runtime_context = runtime_context if runtime_context is not None else RuntimeContext()

    # init frontend
    e_frontend_config = combine_with_default_config(
        ENGINE_FRONTEND_CONFIG, engine_frontend_config)

    frontend_factory_cls = load_class(
        e_frontend_config.component_factory_cls, ConfigurableFactory)
    frontend_factory = frontend_factory_cls(app_config, frontend_cls)
    runtime_context.frontend_thread = OthreadManager(
        e_frontend_config, runtime_context, frontend_factory)
    runtime_context.frontend_thread.setDaemon(True)

    # init transport
    default_engine_transport_config = ENGINE_TRANSPORT_BRIDGE_CONFIG \
        if backend_cls is not None else ENGINE_TRANSPORT_CONFIG

    e_transport_config = engine_transport_config
    if engine_transport_config in (ENGINE_TRANSPORT_CONFIG, ENGINE_TRANSPORT_BRIDGE_CONFIG):
        e_transport_config = None

    e_transport_config = combine_with_default_config(
        default_engine_transport_config, e_transport_config)

    e_backend_config = combine_with_default_config(
        ENGINE_BACKEND_CONFIG, engine_backend_config)

    if transport_cls:
        queue_size = _queue_size(
            engine_frontend_config,
            ENGINE_FRONTEND_CONFIG,
            e_transport_config.thread_num)
        e_frontend_config.queue_size = queue_size
        runtime_context.frontend_thread_queue = Queue.Queue(queue_size)

        if backend_cls:
            queue_size = _queue_size(
                engine_backend_config,
                ENGINE_BACKEND_CONFIG,
                e_backend_config.thread_num)
            queue_size = max(queue_size, e_frontend_config.queue_size)
            e_backend_config.queue_size = queue_size
            runtime_context.backend_thread_queue = Queue.Queue(queue_size)

        transport_factory_cls = load_class(
            e_transport_config.component_factory_cls, ConfigurableFactory)
        transport_factory = transport_factory_cls(app_config, transport_cls)
        runtime_context.transport_thread = OthreadManager(
            e_transport_config, runtime_context, transport_factory)
    else:
        queue_size = _queue_size(
            engine_frontend_config,
            ENGINE_FRONTEND_CONFIG,
            e_backend_config.thread_num)

        e_frontend_config.queue_size = queue_size
        e_backend_config.queue_size = queue_size
        runtime_context.frontend_thread_queue = Queue.Queue(queue_size)

        # init backend
    if backend_cls:
        backend_factory_cls = load_class(
            e_backend_config.component_factory_cls, ConfigurableFactory)
        backend_factory = backend_factory_cls(app_config, backend_cls)
        runtime_context.backend_thread = OthreadManager(
            e_backend_config, runtime_context, backend_factory)

    e_config = combine_with_default_config(ENGINE_CONFIG, engine_config)
    return Engine(e_config, runtime_context)



import pytest

from os_m3_engine.core.engine import RuntimeContext
from os_m3_engine.launcher import create


@pytest.fixture(scope='function')
def config():
    from os_config import Config
    return Config.create()


def test_create_001():
    runtime_context = RuntimeContext()
    _ = create(runtime_context=runtime_context)
    for attr in ['frontend_thread_queue', 'backend_thread_queue',
                 'frontend_thread', 'backend_thread', 'transport_thread']:
        assert hasattr(runtime_context, attr)


def test_create_002():
    with pytest.raises(ValueError):
        create(frontend_cls=None)
    with pytest.raises(ValueError):
        create(transport_cls=None, backend_cls=None)


def test_create_003(config):
    config.queue_size = 1000
    config.thread_num = 10
    runtime_context = RuntimeContext()
    _ = create(engine_frontend_config=config,
               engine_backend_config=config,
               runtime_context=runtime_context)
    assert runtime_context.frontend_thread_queue.maxsize == config.queue_size
    assert runtime_context.frontend_thread.thread_num == config.thread_num
    assert runtime_context.backend_thread_queue.maxsize == config.queue_size
    assert runtime_context.backend_thread.thread_num == config.thread_num


def test_create_004(config):
    with pytest.raises(ImportError):
        create('not_exist.What')


def test_create_005(config):
    runtime_context = RuntimeContext()
    _ = create(transport_cls=None, runtime_context=runtime_context)
    for attr in ['thransport_thread', 'backend_thread_queue']:
        assert not hasattr(runtime_context, attr)

    runtime_context = RuntimeContext()
    _ = create(backend_cls=None, runtime_context=runtime_context)
    for attr in ['backend_thread', 'backend_thread_queue']:
        assert not hasattr(runtime_context, attr)


def test_create_006(config):
    runtime_context = RuntimeContext()
    config.thread_num = 10
    _ = create(engine_frontend_config=config, runtime_context=runtime_context)
    assert runtime_context.frontend_thread_queue.maxsize > 10


def test_create_007(config):
    runtime_context = RuntimeContext()
    config.queue_size = 10
    _ = create(engine_frontend_config=config, runtime_context=runtime_context)
    assert runtime_context.frontend_thread_queue.maxsize == 10


def test_create_008(config):
    runtime_context = RuntimeContext()
    config.thread_num = 10
    _ = create(engine_backend_config=config, runtime_context=runtime_context)
    assert runtime_context.backend_thread_queue.maxsize > 10


def test_create_009(config):
    runtime_context = RuntimeContext()
    config.thread_num = 10
    _ = create(transport_cls=None, engine_backend_config=config,
               runtime_context=runtime_context)
    assert runtime_context.frontend_thread_queue.maxsize > 10


def test_create_010(config):
    runtime_context = RuntimeContext()
    config.thread_num = 10
    _ = create(backend_cls=None, engine_transport_config=config,
               runtime_context=runtime_context)
    assert runtime_context.frontend_thread_queue.maxsize > 10
